"""
MercurySQLite
=============

A Pythonic wrapper for the built-in sqlite3 library.

This package provides classes for interacting with a SQLite database, including:

- DataBase: Represents a SQLite database and provides methods for creating tables, executing SQL commands, and retrieving table objects.

- Table: Represents a table in the SQLite database and provides methods for adding columns, deleting columns, inserting rows, and executing queries.

- Exp: Represents a query expression that can be used to construct complex queries.

Please refer to the individual class documentation for more details.
"""

from .gensql import DataBase, Table, set_driver
from . import drivers



"""
MercurySQL.gensql.exp
=====================
This file offers the `Exp` class, which is used to construct complex query expressions for the MercurySQL package.

Classes
-------
- `BasicExp`: Base class for basic expressions.
- `Exp`: Class for constructing complex query expressions.
"""

from typing import Any, Union, Tuple

from ..errors import *


# ========= Class Decorations =========
class BasicExp:
    pass


class Exp(BasicExp):
    pass


# ========= Classes =========
class BasicExp:
    def __init__(self, exp1, oper="", exp2=None):
        self.exp1 = exp1
        self.oper = oper
        self.exp2 = exp2
        self._formula = ("", ())

        self.gen_formula()

    @staticmethod
    def convert(value: Any) -> Tuple[str, tuple]:
        """
        Convert value into a form that can be used in a SQL query.

        :param value: The value to convert.
        :type value: Any

        :return: The converted value in the form of `(sql_command, paras)`.
        :rtype: Tuple[str, tuple]

        Example Usage:

        .. code-block:: python

            BasicExp.convert(1)                 # ('?', (1,))
            BasicExp.convert('Bernie')          # ('?', ('Bernie',))
            BasicExp.convert(None)              # ('', ())
            BasicExp.convert(Exp('id') == 1)    # ('(id = ?)', (1,))
            ...

        """
        if isinstance(value, BasicExp):
            formula, paras = value.formula()
        elif value is None:
            formula, paras = "", ()
        else:
            formula, paras = "___!!!PAYLOAD!!!___", (value,)

        return formula, paras

    def gen_formula(self) -> None:
        """
        [Helper] Generate the formula of the expression.

        .. note :: You can also construct a formula by yourself, just set the `_formula` attribute to a tuple in the form of `(sql_command, paras)`.
        """
        if self.oper == "":
            self._formula = self.exp1, ()
        else:
            exp1 = BasicExp.convert(self.exp1)
            exp2 = BasicExp.convert(self.exp2)

            self._formula = f"({exp1[0]} {self.oper} {exp2[0]})", tuple(
                exp1[1] + exp2[1]
            )

    def formula(self) -> Tuple[str, tuple]:
        """
        Return the formula of the expression in the form of (sql_command, paras).
        """
        return self._formula


class Exp(BasicExp):
    def __init__(self, o1, op="", o2="", **kwargs):
        """
        Acceptable addition attributes:
        - table: Table ...................... the table to search
        - _str: str ......................... the string to show when print the object
        """
        super().__init__(o1, op, o2)

        self.table = kwargs.get("table", None)
        self._str = kwargs.get("_str", "<MercurySQL.core.Exp object>")
        self.result = None

        if isinstance(o1, Exp):
            self.table = self.table or o1.table
        if isinstance(o2, Exp):
            self.table = self.table or o2.table

    def __eq__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, "=", __value)

    def __ne__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, "<>", __value)

    def __lt__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, "<", __value)

    def __le__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, "<=", __value)

    def __gt__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, ">", __value)

    def __ge__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, ">=", __value)

    def between(
        self, __value1: Union[Exp, int, str], __value2: Union[Exp, int, str]
    ) -> Exp:
        # construct a new Exp object manually: set the `_formula` attribute to `(sql, paras)`.
        res = Exp(self, "BETWEEN", __value1)

        exp1 = res.formula()
        exp3 = BasicExp.convert(__value2)

        # set the `_formula` attribute.
        res._formula = f"({exp1[0]} AND {exp3[0]})", tuple(exp1[1] + exp3[1])

        return res

    def in_(self, __value: Union[list, tuple, set]) -> Exp:
        return Exp(self, "IN", str(tuple(__value)))

    def like(self, __value: str) -> Exp:
        return Exp(self, "LIKE", __value)

    def __and__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, "AND", __value)

    def __or__(self, __value: Union[Exp, int, str]) -> Exp:
        return Exp(self, "OR", __value)

    def __invert__(self) -> Union[Exp, int, str]:
        return Exp("", "NOT", self)

    def query(self, table=None, select="*") -> list:
        """
        Execute query.
        """
        self.table = table or self.table
        self.driver = self.table.db.driver

        if self.table is None:
            raise NotSpecifiedError("Table not specified.")

        condition, paras = self.formula()

        cmd = self.driver.APIs.gensql.query(self.table.table_name, select, condition)
        res = self.table.db.do(cmd, paras=[paras])
        return res.fetchall()

    def __iter__(self):
        """
        use magic method `__iter__` to search.
        """
        if self.result is None:
            if self.table is None:
                raise NotSpecifiedError("Table not specified.")
            
            data = self.query()
            keys = self.table.columns
            self.result = [dict(zip(keys, row)) for row in data]

        return iter(self.result)

    def delete(self, table=None) -> None:
        """
        Execute delete.
        """
        self.table = table or self.table

        if self.table is None:
            raise NotSpecifiedError("Table not specified.")

        condition, paras = self.formula()

        cmd = self.driver.APIs.gensql.delete(self.table.table_name, condition)
        self.table.db.do(cmd, paras=[paras])

    def __str__(self):
        return self._str



from typing import List, Union, Any


class BaseDriver:
    class Cursor:
        pass

    class Conn:
        pass

    class APIs:
        pass


class BaseDriver:
    """
    The base class of all drivers.
    """
    dependencies = []   # just for hints about what dependencies are needed
    version = '0.0.0'
    payload = '?'

    class Cursor:
        """
        .. note::
            The `Cursor` class here is only used for type hints, indicating the return type or parameter type of functions. In practice, instances of the `Cursor` class are not directly created or called.

        A cursor object for database operations.
        Usually, SQL libraries will provide these methods for executing SQL queries and interacting with databases.
        E.g., `sqlite3.Cursor <https://docs.python.org/3/library/sqlite3.html#cursor-objects>`_ or `pymysql.cursors.Cursor <https://pymysql.readthedocs.io/en/latest/modules/cursors.html>`_.
        """

        def execute(self, sql: str, paras: List[tuple] = []) -> BaseDriver.Cursor:
            """
            This method is used to execute SQL queries.

            :param sql: The SQL query to be executed.
            :param paras: The parameters of the SQL query.
            :return: The cursor object itself.
            """
            pass

        def fetchone(self) -> tuple:
            """
            This method is used to fetch the next row of the query result.

            :return: The next row of the query result.
            """
            pass

        def fetchall(self) -> List[tuple]:
            """
            This method is used to fetch all rows of the query result.

            :return: All rows of the query result.
            """
            pass

    class Conn:
        """
        .. note::
            The `Conn` class here is only used for type hints, indicating the return type or parameter type of functions. In practice, instances of the `Cursor` class are not directly created or called.

        A Connection object for database operations.

        Also, a SQL library will provide these methods for interacting with databases.
        E.g., `sqlite3.Connection <https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection>`_ or `pymysql.connections.Connection <https://pymysql.readthedocs.io/en/latest/modules/connections.html>`_.
        """

        def cursor(self) -> BaseDriver.Cursor:
            """
            This method is used to create a cursor object for database operations.

            :return: A cursor object for database operations.
            """
            pass

        def commit(self) -> None:
            """
            This method is used to commit (it means 'save') the changes to the database.
            """
            pass

        def close(self) -> None:
            """
            This method is used to close the connection to the database.
            """
            pass

    class APIs:
        """
        APIs is a class that provides SQL statements for database operations.

        .. warning::
            You must implement all the methods in this class.

        Every definition will be followed by an example of it's return value (in SQLite).
        """
        class gensql:
            """
            APIs in this class will return a SQL code for database operations.
            Generally, these returned codes will be executed **DIRECTLY**.
            """
            @staticmethod
            def drop_table(table_name: str) -> str:
                """
                Drop (Delete) a table.

                :param table_name: The name of the table to be dropped.
                :type table_name: str
                :return: The SQL statement to drop the table.

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"DROP TABLE {table_name}"

                """
                pass
                # return f"DROP TABLE {table_name}"

            @staticmethod
            def get_all_tables() -> str:
                """
                Get all table's informations in the database.

                :return: The SQL statement to get all table's informations in the database.

                Example Implementation (SQLite):

                .. code-block:: python

                    return "SELECT name FROM sqlite_master WHERE type='table';"
                """
                pass
                # return "SELECT name FROM sqlite_master WHERE type='table';"

            @staticmethod
            def get_all_columns(table_name: str) -> str:
                """
                Get all column's informations in the table.

                :param table_name: The name of the table.
                :type table_name: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"PRAGMA table_info({table_name});"

                """
                pass
                # return f"PRAGMA table_info({table_name});"

            @staticmethod
            def create_table_if_not_exists(table_name: str, column_name: str, column_type: str, primaryKey=False, autoIncrement=False) -> Union[str, List[str]]:
                """
                Create a table if it does not exist.

                :param table_name: The name of the table to be created.
                :type table_name: str
                :param column_name: The name of the column to be created.
                :type column_name: str
                :param column_type: The type of the column to be created.
                :type column_type: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f\"\"\"
                        CREATE TABLE IF NOT EXISTS {table_name} ({column_name} {column_type} {'PRIMARY KEY' if primaryKey else ''})
                    \"\"\"

                """

                pass
                # return f"""
                #     CREATE TABLE IF NOT EXISTS {table_name} ({column_name} {column_type} {'PRIMARY KEY' if primaryKey else ''})
                # """

            @staticmethod
            def add_column(table_name: str, column_name: str, column_type: str) -> Union[str, List[str]]:
                """
                Add a column to a table.

                :param table_name: The name of the table to be added.
                :type table_name: str
                :param column_name: The name of the column to be added.
                :type column_name: str
                :param column_type: The type of the column to be added (has been parsed by TypeParser).
                :type column_type: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f\"\"\"
                        ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}
                    \"\"\"

                """
                pass
                # return f"""
                #     ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}
                # """

            @staticmethod
            def drop_column(table_name: str, column_name: str) -> Union[str, List[str]]:
                """
                Drop (Delete) a column from a table.

                :param table_name: The name of the target table.
                :type table_name: str
                :param column_name: The name of the column to be dropped.
                :type column_name: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f\"\"\"
                        ALTER TABLE {table_name} DROP COLUMN {column_name}
                    \"\"\"

                """
                pass
                # return f"""
                #     ALTER TABLE {table_name} DROP COLUMN {column_name}
                # """

            @staticmethod
            def set_primary_key(table, keyname: str, keytype: str) -> Union[str, List[str]]:
                """
                Set a primary key for the specified table.

                :param table: The table to be set.
                :type table: Table
                :param keyname: The name of the primary key.
                :type keyname: str
                :param keytype: The type of the primary key.
                :type keytype: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return [
                        f"CREATE TABLE new_table ({keyname} {keytype} PRIMARY KEY, {', '.join([f'{name} {type_}' for name, type_ in table.columns.items() if name != keyname])})",
                        f"INSERT INTO new_table SELECT * FROM {table.table_name}",
                        f"DROP TABLE {table.table_name}",
                        f"ALTER TABLE new_table RENAME TO {table.table_name}"
                    ]

                """
                pass
                # return [
                #     f"CREATE TABLE new_table ({keyname} {keytype} PRIMARY KEY, {', '.join([f'{name} {type_}' for name, type_ in table.columns.items() if name != keyname])})",
                #     f"INSERT INTO new_table SELECT * FROM {table.table_name}",
                #     f"DROP TABLE {table.table_name}",
                #     f"ALTER TABLE new_table RENAME TO {table.table_name}"
                # ]

            @staticmethod
            def insert(table_name: str, columns: str, values: str) -> str:
                """
                Insert a row into the specified table.

                :param table_name: The name of the table to be inserted.
                :type table_name: str
                :param columns: The columns to be inserted.
                :type columns: str
                :param values: The values to be inserted.
                :type values: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"INSERT INTO {table_name} ({columns}) VALUES ({values})"

                """
                pass
                # return f"INSERT INTO {table_name} ({columns}) VALUES ({values})"

            @staticmethod
            def insert_or_update(table_name: str, columns: str, values: str) -> str:
                """
                Insert a row into the specified table or update it if it already exists.

                :param table_name: The name of the table to be inserted.
                :type table_name: str
                :param columns: The columns to be inserted, already been seperated by ','.
                :type columns: str
                :param values: The values to be inserted, already been seperated by ','.
                :type values: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"INSERT OR REPLACE INTO {table_name} ({columns}) VALUES ({values})"

                """
                pass
                # return f"INSERT OR REPLACE INTO {table_name} ({columns}) VALUES ({values})"

            @staticmethod
            def update(table_name: str, columns: str, condition: str) -> str:
                """
                Update the specified table.

                :param table_name: The name of the table to be updated.
                :type table_name: str
                :param columns: The columns to be updated. Values are already in the format of 'column1=value1, column2=value2, ...', seperated by ','.
                :type columns: str
                :param condition: The condition of the update, in the general SQL format, generated by Exp.
                :type condition: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"UPDATE {table_name} SET {columns} WHERE {condition}"

                """
                pass
                # return f"UPDATE {table_name} SET {columns} WHERE {condition}"

            @staticmethod
            def query(table_name: str, selection: str, condition: str) -> str:
                """
                Query the specified table.

                :param table_name: The name of the table to be queried.
                :type table_name: str
                :param selection: The columns to be selected, seperated by ','.
                :type selection: str
                :param condition: The condition of the query, in the general SQL format, generated by Exp.
                :type condition: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"SELECT {selection} FROM {table_name} WHERE {condition}"

                """
                pass
                # return f"SELECT {selection} FROM {table_name} WHERE {condition}"

            @staticmethod
            def delete(table_name: str, condition: str) -> str:
                """
                Delete the rows in specified table, which matches the condition.

                :param table_name: The name of the table to be deleted.
                :type table_name: str
                :param condition: The condition of the delete, in the general SQL format, generated by Exp.
                :type condition: str

                Example Implementation (SQLite):

                .. code-block:: python

                    return f"DELETE FROM {table_name} WHERE {condition}"
                """
                pass
                # return f"DELETE FROM {table_name} WHERE {condition}"

        @classmethod
        def get_all_tables(cls, conn: BaseDriver.Conn) -> List[str]:
            """
            Get all table's informations in the database.

            .. note:
                You can copy the implementation of this method directly.

            The default implementation is based on the `cls.gensql.get_all_tables()` method.

            :param conn: The connection object of the database.
            :type conn: BaseDriver.Conn

            :return: All table's informations in the database
            """
            cursor = conn.cursor()
            cursor.execute(cls.gensql.get_all_tables())
            return list(map(lambda x: x[0], cursor.fetchall()))

        @classmethod
        def get_all_columns(cls, conn: BaseDriver.Conn, table_name: str) -> List[str]:
            """
            Get all column's informations in the table.

            .. note:
                You can copy the implementation of this method directly.

            The default implementation is based on the `cls.gensql.get_all_columns(table_name)` method.

            :param conn: The connection object of the database.
            :type conn: BaseDriver.Conn
            :param table_name: The name of the table.
            :type table_name: str

            :return: All column's informations in the table.
            :rtype: List[str]. Each element is a list of `[column_name, column_type]`.
            """
            cursor = conn.cursor()
            cursor.execute(cls.gensql.get_all_columns(table_name))
            return cursor.fetchall()

    class TypeParser:
        """
        Parse the type from `Python Type` -> `SQL Type`.
        """

        @staticmethod
        def parse(type_: Any) -> str:
            """
            Compile the type to SQLite type.

            :param type_: The type to parse.
            :type type_: Any

            :return: The SQLite type.
            :rtype: str

            .. note::
                I've provided a example implementation of this method, for SQLite.
                Look at the source code.

            Example Usage:

            .. code-block:: python

                TypeParser.parse(str)       # TEXT
                TypeParser.parse(int)       # INTEGER
                TypeParser.parse(float)     # REAL
                ...

            """
            supported_types = {
                str: 'TEXT',
                int: 'INTEGER',
                float: 'REAL',
                bool: 'BOOLEAN',
                bytes: 'BLOB'
            }

            # round 1: Built-in Types
            if type_ in supported_types:
                return supported_types[type_]

            # round 2: 'Not Null' Types
            """ 
            === !!! Not Supported By SQLite !!! ===

            for t in supported_types:
                if isinstance(type_, t):
                    if type_ == t(not None):
                        return supported_types[t] + ' NOT NULL'
            """

            # round 3: Custom Types
            if isinstance(type_, str):    # custom type
                return type_

            # Not Supported
            raise Exception(f"Type `{str(type_)}` not supported.")

    @staticmethod
    def connect(db_name: str, **kwargs) -> BaseDriver.Conn:
        """
        Connect to the database.

        :param db_name: The name of the database to connect.
        :type db_name: str
        :param kwargs: The parameters of the connection. E.g., `host`, `port`, `user`, `password`, ...
        
        :return: The connection object of the database.
        :rtype: BaseDriver.Conn
        """
        pass



"""
MercurySQL.gensql.table
=======================
This file contains the `Table` class and related classes for handling tables & queries in the SQL database.

Classes
-------
- `Table`: Represents a table in the SQL database and provides methods for adding columns, deleting columns, inserting rows, and executing queries.
- `QueryResult`: Represents the result of a query and provides methods for accessing the query results.
"""

from typing import Any
from ..errors import *
from .exp import Exp


# ========== Class Decorations ==========
class Table:
    pass


class QueryResult:
    class QueryResultRow:
        pass


# ========== Classes ==========
class Table:
    """
    Represents a table in the SQL database and provides methods for adding columns, deleting columns, inserting rows, and executing queries.
    """

    def __init__(self, db, table_name: str):
        """
        Initialize a table object.

        :param db: The database object.
        :type db: DataBase
        :param table_name: The name of the table.
        :type table_name: str

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            table = db['test']

        How It Works:
            - gather all infomations of the table, including columns name and their types.
            - set `isEmpty` to True if the table doesn't have any columns. This variable will effect how the `newColumn()` method works.
        """
        self.db = db
        self.table_name = table_name
        self.driver = db.driver

        all_tables = self.driver.APIs.get_all_tables(self.db)
        self.isEmpty = table_name not in all_tables

        self._gather_info()

    def _gather_info(self):
        """
        [Helper] Gather all infomations of the table, including:
            - `columns`: list[str] .................. name of all columns
            - `columnsType`: dict[str, str] .......... the type of each column
        """
        if not self.isEmpty:
            self.column_info = self.driver.APIs.get_all_columns(
                self.db, self.table_name
            )
        else:
            self.column_info = []

        self.columns = list(map(lambda x: x[0], self.column_info))  # [name, type]
        self.columnsType = {
            self.column_info[i][0]: self.column_info[i][1]
            for i in range(len(self.column_info))
        }

    def __getitem__(self, key: str) -> Exp:
        """
        get a column from the table, mainly used to construct query.

        :param key: The name of the column.
        :type key: str

        :return: An Exp object.

        Example Usage:

        1. Construct a query expression.

        .. code-block:: python

            table = db['test']
            exp = table['id'] == 1
            res = table.select(exp)

        2. Get the definition of a column.

        .. code-block:: python

            table = db['test']
            print(table['id'])  # INTEGER PRIMARY KEY

        How It Works:
            - return an `Exp` object with `table` and `_str` attributes setted.
            - `table` is used to execute the query using `list(...)`, even if table not specified. **[NOT RECOMMENDED]**
            - `_str` is used to print the definition of a column using `str(...)`.
            - raise an Exception if column not exists.
        """
        if key not in self.columns:
            raise NotExistsError(f"Column `{key}` not exists.")

        return Exp(key, table=self, _str=self.columnsType[key])

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Create a new column in the table.

        :param key: The name of the column.
        :type key: str
        :param value: The type of the column.
        :type value: Any. Can be single un-parsed type (e.g. `int`, `str(5)`, ...) OR a tuple in the form of `(type, options)`. (e.g. `(int, 'primary key', 'auto increment')`)

        Options in `value`:
        - 'primary key': Set the column as the primary key of the table.
        - 'auto increment': Set the column as an auto-incremented column.

        Example Usage:

        .. code-block:: python

            table = db['test']
            table['name'] = str
            table['id'] = int, 'primary key', 'auto increment'

        How It Works:
            - get options from `value` if it has parameters (Judge it by whether it's a tuple, so you can use it as the L3 of example showed).
            - Actually create the column, using `newColumn()` method.
        """
        options = {
            "primary key": False,
            "auto increment": False,
        }

        if isinstance(value, tuple):
            # (type, options)
            for i in range(1, len(value)):
                options[value[i].lower()] = True
            value = value[0]

        self.newColumn(
            key,
            value,
            primaryKey=options["primary key"],
            autoIncrement=options["auto increment"],
            force=True,
        )

    def __delitem__(self, key: str) -> None:
        """
        Delete an existing column in this table.
        An Exception will be raised if column not exist.
        Same as `delColumn()`.

        :param key: The name of the column.
        :type key: str

        Example Usage:

        .. code-block:: python

            table = db['test']
            del table['name']  # same as table.delColumn('name')

        """
        self.delColumn(key)

    def select(self, exp: Exp = None, selection: str = "*") -> list:
        """
        Select data from the table.

        :param exp: The query expression.
        :type exp: Exp
        :param selection: The columns to select, default is '*'(all columns).
        :type selection: str

        :return: A list of data.
        :rtype: list

        Example Usage:

        .. code-block:: python

            table = db['test']
            table.select(table['id'] == 1)  # select all columns where id = 1

        How It Works:
            - Construct a `QueryResult` object, which will execute the query whthin it's `QueryResult.__init__()` method.
            - return a `QueryResult` object with results.
        """
        if exp is None:
            exp = Exp(1, "=", 1)

        return QueryResult(self, exp, selection)

    def newColumn(
        self, name: str, type_: Any, force=False, primaryKey=False, autoIncrement=False
    ) -> None:
        """
        Add a new column to the table.

        :param name: The name of the column.
        :type name: str
        :param type_: The type of the column.
        :type type_: Any
        :param force: Allow to skip when processing an existing column.
        :type force: bool
        :param primaryKey: Whether the column is a primary key.
        :type primaryKey: bool
        :param autoIncrement: Whether the column is auto-incremented.
        :type autoIncrement: bool

        Example Usage:

        .. code-block:: python

            table = db['test']
            table.newColumn('name', str)
            table.newColumn('id', int, primaryKey=True)

        How It Works:
            - Create a new table with this column, if table is empty.
            - Add the column to an existing table.
            - Set as the `primary key` column if `primaryKey` is True.
            - Record its name and type in `self.columns` and `self.columnsType`.
        """
        if name in self.columns:
            if not force:
                raise DuplicateError(f"Column `{name}` already exists.")
            else:
                return

        type_ = self.driver.TypeParser.parse(type_)

        if self.isEmpty:
            # create it first
            cmd = self.driver.APIs.gensql.create_table_if_not_exists(
                self.table_name,
                name,
                type_,
                primaryKey=primaryKey,
                autoIncrement=autoIncrement,
            )
            self.db.do(cmd)
            self.isEmpty = False
        else:
            cmd = self.driver.APIs.gensql.add_column(self.table_name, name, type_)
            self.db.do(cmd)

            if primaryKey:
                self.setPrimaryKey(name, type_)

        self.columns.append(name)
        self.columnsType[name] = type_

    def struct(
        self, columns: dict, force=True, primaryKey: str = None, autoIncrement=False
    ) -> None:
        """
        Set the structure of the table.

        :param columns: The structure of the table.
        :type columns: dict
        :param force: Allow to skip when column exist and have the same type at the same time.
        :type force: bool
        :param primaryKey: The primary key of the table.
        :type primaryKey: str

        Example Usage:

        .. code-block:: python

            table = db['test']
            table.struct({
                'id': int,
                'name': str
            }, primaryKey='id')

        """
        for name, type_ in columns.items():
            type_origin = type_
            type_ = self.driver.TypeParser.parse(type_)
            isPrimaryKey = name == primaryKey

            if name in self.columns:
                if type_.lower() != self.columnsType[name].lower():
                    raise ConfilictError(
                        f"Column `{name}` with different types (`{self.columnsType[name]}`) already exists. While trying to add column `{name}` with type `{type_}`."
                    )
                elif not force:
                    raise DuplicateError(
                        f"Column `{name}` already exists. You can use `force=True` to avoid this error."
                    )
            else:
                self.newColumn(
                    name,
                    type_origin,
                    primaryKey=isPrimaryKey,
                    autoIncrement=autoIncrement,
                )

    def delColumn(self, name: str) -> None:
        if name not in self.columns:
            # column not exist
            raise NotExistsError(f"Column `{name}` not exist!")
        elif len(self.columns) == 1:
            # the last column, delete the table instead
            self.db.deleteTable(self.table_name)
        else:
            # delete the column
            self.columns.remove(name)
            cmd = self.driver.APIs.gensql.drop_column(name)
            self.db.do(cmd)

    def setPrimaryKey(self, keyname: str, keytype: str) -> None:
        """
        Set a column as the primary key of the table.

        :param keyname: The name of the column.
        :type keyname: str
        :param keytype: The type of the column, been parsed by `TypeParser`.
        :type keytype: str

        Example Usage:

        .. code-block:: python

            table = db['test']
            table.setPrimaryKey('id')

        """
        cmd = self.driver.APIs.gensql.set_primary_key(self, keyname, keytype)
        self.db.do(cmd)

    def insert(self, __auto=False, **kwargs) -> None:
        """
        Insert a row into the table.

        :param \_\_auto: Whether to update the row if it already exists.
        :type \_\_auto: bool
        :param \*\*kwargs: The data to insert.

        Example Usage:

        .. code-block:: python

            table = db['test']
            table.insert(id=1, name='Bernie', age=15, __auto=True)

        """
        # get keys and clean them
        keys = list(kwargs.keys())
        if "__auto" in keys:
            __auto = kwargs["__auto"]
            keys.remove("__auto")

        columns = ", ".join(keys)
        values = ", ".join([self.driver.payload for _ in range(len(keys))])

        # Determine the SQL command based on the value of `__auto`
        if __auto:
            cmd = self.driver.APIs.gensql.insert_or_update(
                self.table_name, columns, values
            )
        else:
            cmd = self.driver.APIs.gensql.insert(self.table_name, columns, values)

        self.db.do(cmd, paras=[tuple(kwargs[k] for k in keys)])

    def update(self, exp: Exp, **kwargs) -> None:
        """
        Update the table.

        :param exp: The query expression.
        :type exp: Exp
        :param \*\*kwargs: The data to update.

        Example Usage:

        .. code-block:: python

            table = db['test']
            table.update(table['id'] == 1, name='Bernie', age=15)

        """
        columns = ", ".join([f"{key} = {self.driver.payload}" for key in kwargs.keys()])
        values = tuple(kwargs.values())

        condition, paras = exp.formula()

        cmd = self.driver.APIs.gensql.update(self.table_name, columns, condition)
        self.db.do(cmd, paras=[values + paras])


class QueryResult:
    """
    [Helper Class]
    The result of a query.

    Example Usage:

    .. code-block:: python

        res = table.select(exp)
        for row in res:
            print(row)

    """

    class QueryResultRow:
        """
        [Helper Class]
        Representing a single row of the query result.
        You can use QueryResult[i] to get it, and use it by `row.column_name`.

        Example Usage:

        .. code-block:: python

            res = table.select(exp)
            row = res[0]
            print(row.id)

        """

        def __init__(self, data: dict):
            self.data = data

        def __getattribute__(self, __name: str) -> Any:
            """
            The main method of this class, used to get the value of a column.

            :param __name: The name of the column.
            :type __name: str

            :return: The value of the column.
            """
            data = object.__getattribute__(self, "data")
            if __name == "data":
                return data
            else:
                return data[__name]

        def __iter__(self):
            """
            Make it possible to iterate through.
            It will work as a `dict`.
            """
            return iter(self.data)

    def __init__(self, table: Table, exp: Exp, selection: str = "*"):
        values = exp.query(table, selection)
        keys = (
            table.columns
            if selection == "*"
            else list(map(lambda x: x.strip(), selection.split(",")))
        )
        self.data = [dict(zip(keys, row)) for row in values]

    def __getitem__(self, index: int) -> Any:
        return self.QueryResultRow(self.data[index])

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)



"""
Requirements: 
  - mysql-connector-python
"""
from .base import BaseDriver

import mysql.connector
from typing import Any, List


class Driver_MySQL(BaseDriver):
    pass


class Driver_MySQL(BaseDriver):
    """
    .. note::
        Supported MySQL Version: **8.2**
    """
    dependencies = ['mysql-connector-python']
    version = '0.1.0'
    payload = '%s'

    Conn = mysql.connector.MySQLConnection
    Cursor = mysql.connector.cursor_cext.CMySQLCursor

    class APIs:
        class gensql:
            @staticmethod
            def drop_table(table_name: str) -> str:
                return f"DROP TABLE {table_name};"

            @staticmethod
            def get_all_tables() -> str:
                return "SHOW TABLES;"

            @staticmethod
            def get_all_columns(table_name: str) -> str:
                return f"DESCRIBE `{table_name}`;"

            @staticmethod
            def create_table_if_not_exists(table_name: str, column_name: str, column_type: str, primaryKey=False, autoIncrement=False, engine='', charset='') -> str:
                return f"""
                    CREATE TABLE IF NOT EXISTS `{table_name}` (
                        `{column_name}` {column_type} {'PRIMARY KEY' if primaryKey else ''} {'AUTO_INCREMENT' if autoIncrement else ''}
                    ) {engine} {f'DEFAULT CHARSET={charset}' if charset else ''};
                """

            @staticmethod
            def add_column(table_name: str, column_name: str, column_type: str) -> str:
                return f"ALTER TABLE `{table_name}` ADD COLUMN `{column_name}` {column_type};"

            @staticmethod
            def drop_column(table_name: str, column_name: str) -> str:
                return f"ALTER TABLE `{table_name}` DROP COLUMN `{column_name}`;"

            @staticmethod
            def set_primary_key(table, keyname: str, keytype: str) -> list:
                return [
                    f"ALTER TABLE `{table.table_name}` DROP PRIMARY KEY;",
                    f"ALTER TABLE `{table.table_name}` ADD PRIMARY KEY (`{keyname}`);"
                ]

            @staticmethod
            def insert(table_name: str, columns: str, values: str) -> str:
                return f"INSERT INTO `{table_name}` ({columns}) VALUES ({values});"
            
            @staticmethod
            def insert_or_update(table_name: str, columns: str, values: str) -> str:
                update_columns = ', '.join(f'{col}=VALUES({col})' for col in columns.split(', '))
                return f"INSERT INTO `{table_name}` ({columns}) VALUES ({values}) ON DUPLICATE KEY UPDATE {update_columns};"

            @staticmethod
            def update(table_name: str, columns: str, condition: str) -> str:
                return f"UPDATE `{table_name}` SET {columns} WHERE {condition};"

            @staticmethod
            def query(table_name: str, selection: str, condition: str) -> str:
                return f"SELECT {selection} FROM `{table_name}` WHERE {condition};"

            @staticmethod
            def delete(table_name: str, condition: str) -> str:
                return f"DELETE FROM `{table_name}` WHERE {condition};"

        @classmethod
        def get_all_tables(cls, conn) -> List[str]:
            cursor = conn.cursor()
            cursor.execute(cls.gensql.get_all_tables())
            return list(map(lambda x: x[0], cursor.fetchall()))

        @classmethod
        def get_all_columns(cls, conn, table_name: str) -> List[str]:
            cursor = conn.cursor()
            cursor.execute(cls.gensql.get_all_columns(table_name))
            # [name, type, null, key, default, extra]
            return list(map(lambda x: [x[0], x[1], x[2], x[3], x[4], x[5]], cursor.fetchall()))

    class TypeParser:
        """
        Parse the type from `Python Type` -> `MySQL Type`.
        """

        @staticmethod
        def parse(type_: Any) -> str:
            """
            Compile the type to MySQL type.

            :param type_: The type to parse.
            :type type_: Any

            :return: The MySQL type.
            :rtype: str

            +-------------------+-------------+
            | Supported Types   | SQLite Type |
            +===================+=============+
            | bool              | BOOLEAN     |
            +-------------------+-------------+
            | int               | INTEGER     |
            +-------------------+-------------+
            | float             | FLOAT       |
            +-------------------+-------------+
            | str               | VARCHAR(225)|
            +-------------------+-------------+
            | bytes             | BLOB        |
            +-------------------+-------------+
            | datetime.datetime | DATETIME    |
            +-------------------+-------------+
            | datetime.date     | DATE        |
            +-------------------+-------------+
            | datetime.time     | TIME        |
            +-------------------+-------------+
            

            Example Usage:

            .. code-block:: python

                TypeParser.parse(int(10))      # INT DEFAULT 10
                TypeParser.parse(str)          # VARCHAR
                TypeParser.parse(float(1.3))   # FLOAT DEFAULT 1.3
                TypeParser.parse('\t DOUBLE DEFAULT 1.23')    # DOUBLE DEFAULT 1.23
                ...

            """
            import datetime

            supported_types = {
                bool: 'BOOLEAN',
                int: 'INT',
                float: 'FLOAT',
                str: 'VARCHAR(225)',
                bytes: 'BLOB',

                datetime.datetime: 'DATETIME',
                datetime.date: 'DATE',
                datetime.time: 'TIME',
            }

            if not isinstance(type_, tuple):
                type_ = (type_,)

            res = ""

            # round 1: Built-in Types
            if type_[0] in supported_types:
                res = supported_types[type_[0]]

            # round 2: Custom Types
            if res == "" and isinstance(type_[0], str) and type_[0].startswith('\t'):    # custom type
                return type_[0].strip()
            
            # round 3: Built-in Types With Default Value
            if type(type_[0]) != type:
                if isinstance(type_[0], bytes):
                    type_[0] = type_[0].decode()
                res = f"{supported_types[type(type_[0])]} DEFAULT {type_[0]}"

            # round 4: Not Null
            for i in range(1, len(type_)):
                if type_[i] == 'not null':
                    res += ' NOT NULL'
                else:
                    res += f" DFAULT {type_[i]}"

            # Not Supported
            if res == "":
                raise TypeError(f"Type `{str(type_)}` is not supported.")
            
            return res

    @staticmethod
    def connect(db_name: str, host: str, user: str, passwd: str = '', force=False) -> Conn:
        """
        Connect to a MySQL database.
        """
        if force:
            return mysql.connector.connect(
                host=host,
                user=user,
                passwd=passwd,
                database=db_name
            )
        else:
            conn = mysql.connector.connect(
                host=host,
                user=user,
                passwd=passwd
            )
            conn.backup_cursor = conn.cursor

            def cursor():
                """ Create Database if not exists. """
                conn.cursor = conn.backup_cursor

                c = conn.cursor()

                try:
                    c.execute(f'USE {db_name};')
                except mysql.connector.errors.ProgrammingError:
                    c.execute(f'CREATE DATABASE {db_name};')
                    c.execute(f'USE {db_name};')

                return c

            conn.cursor = cursor
            return conn



"""
Requirements: 
  - sqlite3
"""
from .base import BaseDriver

import sqlite3
from typing import Any, List


class Driver_SQLite(BaseDriver):
    pass


class Driver_SQLite(BaseDriver):
    dependencies = ['sqlite3']
    version = '0.1.0'
    payload = '?'

    Conn = sqlite3.Connection
    Cursor = sqlite3.Cursor

    class APIs:
        class gensql:
            @staticmethod
            def drop_table(table_name: str) -> str:
                return f"DROP TABLE {table_name}"

            @staticmethod
            def get_all_tables() -> str:
                return "SELECT name FROM sqlite_master WHERE type='table';"

            @staticmethod
            def get_all_columns(table_name: str) -> str:
                return f"PRAGMA table_info({table_name});"

            @staticmethod
            def create_table_if_not_exists(table_name: str, column_name: str, column_type: str, primaryKey=False, autoIncrement=False) -> str:
                return f"""
                    CREATE TABLE IF NOT EXISTS {table_name} ({column_name} {column_type} {'PRIMARY KEY' if primaryKey else ''} {'AUTOINCREMENT' if autoIncrement else ''})
                """

            @staticmethod
            def add_column(table_name: str, column_name: str, column_type: str) -> str:
                return f"""
                    ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}
                """

            @staticmethod
            def drop_column(table_name: str, column_name: str) -> str:
                return f"""
                    ALTER TABLE {table_name} DROP COLUMN {column_name}
                """

            @staticmethod
            def set_primary_key(table, keyname: str, keytype: str) -> list:
                return [
                    f"CREATE TABLE ___temp_table ({keyname} {keytype} PRIMARY KEY, {', '.join([f'{name} {type_}' for name, type_ in table.columnsType.items() if name != keyname])})",
                    f"INSERT INTO ___temp_table SELECT * FROM {table.table_name}",
                    f"DROP TABLE {table.table_name}",
                    f"ALTER TABLE ___temp_table RENAME TO {table.table_name}"
                ]

            @staticmethod
            def insert(table_name: str, columns: str, values: str) -> str:
                return f"INSERT INTO {table_name} ({columns}) VALUES ({values})"

            @staticmethod
            def insert_or_update(table_name: str, columns: str, values: str) -> str:
                return f"INSERT OR REPLACE INTO {table_name} ({columns}) VALUES ({values})"

            @staticmethod
            def update(table_name: str, columns: str, condition: str) -> str:
                return f"UPDATE {table_name} SET {columns} WHERE {condition}"

            @staticmethod
            def query(table_name: str, selection: str, condition: str) -> str:
                return f"SELECT {selection} FROM {table_name} WHERE {condition}"

            @staticmethod
            def delete(table_name: str, condition: str) -> str:
                return f"DELETE FROM {table_name} WHERE {condition}"

        @classmethod
        def get_all_tables(cls, db) -> List[str]:
            cursor = db.do(cls.gensql.get_all_tables())
            return list(map(lambda x: x[0], cursor.fetchall()))

        @classmethod
        def get_all_columns(cls, db, table_name: str) -> List[str]:
            cursor = db.do(cls.gensql.get_all_columns(table_name))
            return list(map(lambda x: [x[1], x[2]], cursor.fetchall()))

    class TypeParser:
        """
        Parse the type from `Python Type` -> `SQL Type`.
        """

        @staticmethod
        def parse(type_: Any) -> str:
            """
            Compile the type to SQLite type.

            :param type_: The type to parse.
            :type type_: Any

            :return: The SQLite type.
            :rtype: str

            +----------------+-------------+
            | Supported Types| SQLite Type |
            +================+=============+
            | str            | TEXT        |
            +----------------+-------------+
            | int            | INTEGER     |
            +----------------+-------------+
            | float          | REAL        |
            +----------------+-------------+
            | bool           | BOOLEAN     |
            +----------------+-------------+
            | bytes          | BLOB        |
            +----------------+-------------+

            Example Usage:

            .. code-block:: python

                TypeParser.parse(str)       # TEXT
                TypeParser.parse(int)       # INTEGER
                TypeParser.parse(float)     # REAL
                ...

            """
            supported_types = {
                str: 'TEXT',
                int: 'INTEGER',
                float: 'REAL',
                bool: 'BOOLEAN',
                bytes: 'BLOB'
            }

            # round 1: Built-in Types
            if type_ in supported_types:
                return supported_types[type_]

            # round 2: 'Not Null' Types
            """ 
            === !!! Not Supported By SQLite !!! ===

            for t in supported_types:
                if isinstance(type_, t):
                    if type_ == t(not None):
                        return supported_types[t] + ' NOT NULL'
            """

            # round 3: Custom Types
            if isinstance(type_, str):    # custom type
                return type_

            # Not Supported
            raise TypeError(f"Type `{str(type_)}` not supported.")

    @staticmethod
    def connect(db_name: str, **kwargs) -> Driver_SQLite.Conn:
        return sqlite3.connect(db_name, **kwargs)
